【问题标题】:How to do weighted average of two device_vectors with points selected using map in thrust?如何使用推力地图选择两个设备向量的加权平均值?
【发布时间】:2018-11-23 13:19:41
【问题描述】:

我有两个 device_vector P & Q(比如大小为 100)。 我有两个用于 P&Q 的 device_vector 映射(大小为 10 的 MapP 和 MapQ),其中包含要从 P&Q 中选择的点的索引。 我有一个 device_vector D 来表示重量。

我需要为使用相应地图选择的 P & Q 中的所有点计算 (P*D+Q)/(D+1)。

我的方法如下。它可以工作,但是太麻烦了。 谁能提出更好的方法?

#include <thrust/device_vector.h>
#include <thrust/random.h>
#include <thrust/sequence.h>
#include <thrust/execution_policy.h>
#include <thrust/transform.h>
#include <thrust/functional.h>
#include <thrust/iterator/transform_iterator.h>
#include <thrust/iterator/permutation_iterator.h>

thrust::device_vector<float> random_vector(const size_t N, 
                                         unsigned int seed = thrust::default_random_engine::default_seed)
{
    thrust::default_random_engine rng(seed);
    thrust::uniform_real_distribution<float> u01(0.0f, 10.0f);
    thrust::device_vector<float> temp(N);
    for(size_t i = 0; i < N; i++) {
        temp[i] = u01(rng);
    }
    return temp;
}

// note: functor inherits from unary_function
struct increment : public thrust::unary_function<int,int>
{
  __host__ __device__
  int operator()(int x) const
  {
    return x + 1;
  }
};

int main(int argc, char * argv[])
{

int N=atoi(argv[1]);

thrust::device_vector<float> P = random_vector(N,1);
thrust::device_vector<float> Q = random_vector(N,9);

thrust::device_vector<int> D(N);
thrust::sequence(thrust::device, D.begin(), D.begin() + N, 1);

thrust::device_vector<float> temp(10);

thrust::device_vector<int> MapP(10); // map
thrust::device_vector<int> MapQ(10); // map

MapP[0]=0;MapP[1]=5;MapP[2]=4;MapP[3]=2;MapP[4]=7;MapP[5]=1;MapP[6]=9;MapP[7]=3;MapP[8]=6;MapP[9]=8;
MapQ[0]=10;MapQ[1]=15;MapQ[2]=12;MapQ[3]=14;MapQ[4]=11;MapQ[5]=17;MapQ[6]=13;MapQ[7]=19;MapQ[8]=18;MapQ[9]=16;


// The weighted average is (D*P+Q)/(D+1)
// We compute D*P first

//thrust::transform(thrust::device, P.begin(), P.end(), D.begin(), temp.begin(), thrust::multiplies<float>()); // use permutation iterator

thrust::transform(thrust::device, thrust::make_permutation_iterator(P.begin(),MapP.begin()),
                                  thrust::make_permutation_iterator(P.end(),MapP.end()),
                  thrust::make_permutation_iterator(D.begin(),MapP.begin()), 
                  temp.begin(), thrust::multiplies<float>());


// Then we add D*p to Q

//thrust::transform(thrust::device, temp.begin(), temp.end(), Q.begin(), temp.begin(), thrust::plus<float>()); // use permutation iterator

thrust::transform(thrust::device, temp.begin(), temp.end(),
                  thrust::make_permutation_iterator(Q.begin(),MapQ.begin()), 
                  temp.begin(), thrust::plus<float>());


// Then we divide by D+1

//thrust::transform(thrust::device, temp.begin(), temp.end(), thrust::make_transform_iterator(D.begin(), increment()), temp.begin(),  thrust::divides<float>());

thrust::transform(thrust::device, temp.begin(), temp.end(),
                  thrust::make_permutation_iterator(D.begin(),MapP.begin()), 
                  temp.begin(), thrust::divides<float>());


// replace contents of P with the weighted sum using pts in map M

thrust::copy(thrust::device, temp.begin(), temp.end(), thrust::make_permutation_iterator(P.begin(),MapP.begin())); // use permutation iterator

return 0;
}

【问题讨论】:

  • (P*D+Q)/(D+1) 在数学上是什么意思?这是一个元素操作吗?如果 P 和 D 是向量,则 P*D 是标量还是矩阵,具体取决于 * 是表示内积还是外积。并且没有定义部门......
  • 元素操作。不是矢量运算。

标签: cuda thrust weighted-average


【解决方案1】:

我假设您希望对向量进行逐元素操作,因为这是您提供的演示代码的行为。

请注意,当传递置换迭代器的结尾时,我们不使用源向量的结尾:

thrust::make_permutation_iterator(P.end(),MapP.end()),
                                  ^^^^^

而是开始:

thrust::make_permutation_iterator(P.begin(),MapP.end()),

有关此示例,请参阅thrust quick start guide

另请注意,在您的问题和代码中,您指的是除以 D+1,但您的代码实际上是除以 D,而不是 D+1。

关于您的问题,一切都可以通过使用适当定义的函子一次调用 thrust::transform 来完成。由于在这个实现中我们需要将多个向量传递给thrust::transform,所以我们将引入thrust::zip_iterator

$ cat t332.cu
#include <thrust/device_vector.h>
#include <thrust/random.h>
#include <thrust/sequence.h>
#include <thrust/execution_policy.h>
#include <thrust/transform.h>
#include <thrust/functional.h>
#include <thrust/iterator/transform_iterator.h>
#include <thrust/iterator/permutation_iterator.h>
#include <thrust/iterator/zip_iterator.h>

thrust::device_vector<float> random_vector(const size_t N,
                                         unsigned int seed = thrust::default_random_engine::default_seed)
{
    thrust::default_random_engine rng(seed);
    thrust::uniform_real_distribution<float> u01(0.0f, 10.0f);
    thrust::device_vector<float> temp(N);
    for(size_t i = 0; i < N; i++) {
        temp[i] = u01(rng);
    }
    return temp;
}

// The weighted average is (D*P+Q)/(D+1)
struct w_avg
{
template <typename T>
  __host__ __device__
  float operator()(T x) const
  {
    return (thrust::get<0>(x)*thrust::get<1>(x)+thrust::get<2>(x))/(thrust::get<1>(x)+1.0f);
  }
};

int main(int argc, char * argv[])
{

int N=atoi(argv[1]);

thrust::device_vector<float> P = random_vector(N,1);
thrust::device_vector<float> Q = random_vector(N,9);

thrust::device_vector<int> D(N);
thrust::sequence(thrust::device, D.begin(), D.begin() + N, 1);


thrust::device_vector<int> MapP(10); // map
thrust::device_vector<int> MapQ(10); // map

MapP[0]=0;MapP[1]=5;MapP[2]=4;MapP[3]=2;MapP[4]=7;MapP[5]=1;MapP[6]=9;MapP[7]=3;MapP[8]=6;MapP[9]=8;
MapQ[0]=10;MapQ[1]=15;MapQ[2]=12;MapQ[3]=14;MapQ[4]=11;MapQ[5]=17;MapQ[6]=13;MapQ[7]=19;MapQ[8]=18;MapQ[9]=16;


// The weighted average is (D*P+Q)/(D+1)

thrust::transform(thrust::device, thrust::make_zip_iterator(thrust::make_tuple(
                                                            thrust::make_permutation_iterator(P.begin(),MapP.begin()),
                                                            thrust::make_permutation_iterator(D.begin(),MapP.begin()),
                                                            thrust::make_permutation_iterator(Q.begin(),MapQ.begin()))),
                                  thrust::make_zip_iterator(thrust::make_tuple(
                                                            thrust::make_permutation_iterator(P.begin(),MapP.end()),
                                                            thrust::make_permutation_iterator(D.begin(),MapP.end()),
                                                            thrust::make_permutation_iterator(Q.begin(),MapQ.end()))),
                                                            thrust::make_permutation_iterator(P.begin(),MapP.begin()),
                                  w_avg());


for (int i = 0; i < 5; i++) {
  std::cout << P[i] << std::endl;}
return 0;
}
$ nvcc -o t332 t332.cu
$ ./t332 100
4.02976
3.75275
5.32832
8.53189
8.46641
$

请注意,上述代码中的仿函数除以 D+1。将其更改为除以 D 将是微不足道的,而是匹配您的代码(但不是您声明的意图)。

【讨论】:

    猜你喜欢
    • 2019-02-19
    • 2016-12-08
    • 2018-10-04
    • 1970-01-01
    • 1970-01-01
    • 1970-01-01
    • 2021-11-03
    • 1970-01-01
    • 2023-03-10
    相关资源
    最近更新 更多